/************************************************************************
 *	Copyright 1989, University Corporation for Atmospheric Research *
 *		Not for Resale. All copies to include this notice.      *
 *                                                                      *
 ************************************************************************/

/*-
OVERVIEW
   This file contains jacket routines written in C for interfacing
   Fortran netCDF function calls to the actual C binding for the
   NetCDF.  This code is written explicitly for UNIX.  In general,
   these functions convert between column-major-order arrays and
   row-major-order arrays, and map between array indicing
   beginning at one and array indicing beginning at zero.
 -*/

/* LINTLIBRARY */
#include	"netcdf.h"
#include	<ctype.h>
#include        <string.h>
#include	<stdio.h>

/* global integer used for suppressing error messages and determining
 * the fatality of errors.
 */
extern int      ncopts;		/* default is (NC_FATAL | NC_VERBOSE) */

/* global integer that contains a netCDF-specific error code */
extern int      ncerr;


/* used to set the C global variable ncopts from Fortran */
int
ncpopt_(val)
    int            *val;  /* NC_FATAL, NC_VERBOSE, or NC_FATAL|NC_VERBOSE */
{
    ncopts = *val;
}


/* used to get the C global variable ncopts from Fortran */
int
ncgopt_(val)
    int            *val;  /* NC_FATAL, NC_VERBOSE, or NC_FATAL|NC_VERBOSE */
{
    *val = ncopts;
}


/* For backward compatibility only.  Use ncpopt() instead. */
int
stncopts_(val)
    int            *val;  /* NC_FATAL, NC_VERBOSE, or NC_FATAL|NC_VERBOSE */
{
    ncopts = *val;
}


static void
reverse(array, length)
     int             array[];	/* array to be reversed */
     int             length;	/* length of array */
{
    int             temp, i, j;

    for (i = 0, j = length - 1; i < j; i++, j--) {
	temp = array[i];
	array[i] = array[j];
	array[j] = temp;
    }
}


/* error handling function */
static void
handle_err(pname, rcode)
    char           *pname;	/* procedure name */
    int             rcode;	/* error return */
{
    if (rcode == 0)
	return;
    if (ncopts & NC_VERBOSE) {
	switch (rcode) {
	case NC_ESTS:
	    (void) fprintf(stderr,
			   "%s: In FORTRAN interface, string too small\n",
			   pname);
	    break;
	case NC_ENTOOL:
	    (void) fprintf(stderr,
			   "%s: length of name exceeds MAX_NC_NAME\n",
			   pname);
	    break;
	default:
	    (void) fprintf(stderr, "%s: unknown error code\n", pname);
	}
    }
    if (ncopts & NC_FATAL)
	exit(1);
}

/* copy function used to copy strings with embedded blanks */
static void
fstrncpy(target, source, maxlen)
    char           *target;	/* space to be copied into */
    char           *source;	/* string to be copied */
    int             maxlen;	/* maximum length of *source */
{
    while (maxlen-- && isprint(*source))
	*target++ = *source++;
    *target = '\0';
}

/* copy function used to copy strings terminated with blanks */
static void
nstrncpy(target, source, maxlen)
    char           *target;	/* space to be copied into */
    char           *source;	/* string to be copied */
    int             maxlen;	/* maximum length of *source */
{
    while (maxlen-- && isgraph(*source))
	*target++ = *source++;
    *target = '\0';
}


/* ------------ UNIX FORTRAN jackets for netCDF Functions ------------ */

/* creates a new netCDF file, returning a netCDF ID.  New netCDF
 * file is placed in define mode.
 */
int
nccre_(pathname, clobmode, rcode, pathlen)
    char           *pathname;	/* file name of new netCDF file */
    int            *clobmode;	/* either NCCLOB or NCNOCLOB */
    int            *rcode;	/* returned error code */
    int             pathlen;
{
    char            name[MAX_NC_NAME + 1];
    int             cdfid;

    (void) nstrncpy(name, pathname, pathlen);
    if ((cdfid = nccreate(name, *clobmode)) != -1) {
	*rcode = 0;
	return (cdfid);
    }
    *rcode = ncerr;
    return (-1);
}


/* opens an existing netCDF file for access */
int
ncopn_(pathname, rwmode, rcode, pathlen)
    char           *pathname;	/* file name for netCDF to be opened */
    int            *rwmode, *rcode;	/* either NCWRITE or NCNOWRIT */
    int             pathlen;	/* returned error code */
{
    char            name[MAX_NC_NAME + 1];
    int             cdfid;

    (void) nstrncpy(name, pathname, pathlen);
    if ((cdfid = ncopen(name, *rwmode)) != -1) {
	*rcode = 0;
	return (cdfid);
    }
    *rcode = ncerr;
    return (-1);
}


/* adds a new dimension to an open netCDF file in define mode */
int
ncddef_(cdfid, dimname, dimlen, rcode, dimnamelen)
    int            *cdfid;	/* netCDF ID */
    char           *dimname;	/* dimension name */
    int            *dimlen;	/* dimension length */
    int            *rcode;	/* returned error code */
    int             dimnamelen;	/* declared length of dimname */
{
    char            name[MAX_NC_NAME + 1];
    int             dimid;

    (void) nstrncpy(name, dimname, dimnamelen);
    if ((dimid = ncdimdef(*cdfid, name, *dimlen)) != -1) {
	*rcode = 0;
	return (dimid + 1);
    }
    *rcode = ncerr;
    return (-1);
}


/* returns the ID of a netCDF dimension, given the name of the
 * dimension
 */
int
ncdid_(cdfid, dimname, rcode, dimnamelen)
    int            *cdfid;	/* netCDF ID */
    char           *dimname;	/* dimension name */
    int            *rcode;	/* returned error code */
    int             dimnamelen;	/* declared length of dimname */
{
    char            name[MAX_NC_NAME + 1];
    int             dimid;

    (void) nstrncpy(name, dimname, dimnamelen);
    if ((dimid = ncdimid(*cdfid, name)) != -1) {
	*rcode = 0;
	return (dimid + 1);
    }
    *rcode = ncerr;
    return (-1);
}


/* adds a new variable to an open netCDF file in define mode */
int
ncvdef_(cdfid, varname, datatype, ndims, dimarray, rcode, varnamelen)
    int            *cdfid;	/* netCDF ID */
    char           *varname;	/* name of variable */
    int            *datatype;	/* netCDF datatype of variable */
    int            *ndims;	/* number of dimensions of variable */
    int            *dimarray;	/* array of ndims dimensions IDs */
    int            *rcode;	/* returned error code */
    int             varnamelen;	/* declared length of varname */
{
    int             varid, i, dimid[MAX_VAR_DIMS];
    char            name[MAX_NC_NAME + 1];

    (void) nstrncpy(name, varname, varnamelen);
    for (i = 0; i < *ndims; i++)
	dimid[i] = dimarray[i] - 1;
    reverse(dimid, *ndims);
    if ((varid = ncvardef(*cdfid, name, (nc_type) *datatype, *ndims,
			  dimid)) != -1) {
	*rcode = 0;
	return (varid + 1);
    }
    *rcode = ncerr;
    return (-1);
}


/* returns the ID of a netCDF variable given its name */
int
ncvid_(cdfid, varname, rcode, varnamelen)
    int            *cdfid;	/* netCDF ID */
    char           *varname;	/* variable name */
    int            *rcode;	/* returned error code */
    int             varnamelen;	/* declared length of varname */
{
    int             varid;
    char            name[MAX_NC_NAME + 1];

    (void) nstrncpy(name, varname, varnamelen);
    if ((varid = ncvarid(*cdfid, name)) != -1) {
	*rcode = 0;
	return (varid + 1);
    }
    *rcode = ncerr;
    return (-1);
}


/* returns number of bytes per netCDF data type */
int
nctlen_(datatype, rcode)
    int            *datatype;	/* netCDF datatype */
    int            *rcode;	/* returned error code */
{
    int itype;
    if ((itype = nctypelen((nc_type) *datatype)) != -1) {
	*rcode = 0;
	return (itype);
    }
    *rcode = ncerr;
    return (-1);
}



/* ------------ UNIX FORTRAN jackets for netCDF Subroutines ---------- */



/* closes an open netCDF file */
void
ncclos_(cdfid, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *rcode;	/* returned error code */
{
    *rcode = 0;
    if (ncclose(*cdfid) == -1)
      *rcode = ncerr;
}

/* puts an open netCDF into define mode */
void
ncredf_(cdfid, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *rcode;	/* returned error code */
{
    *rcode = 0;
    if (ncredef(*cdfid) == -1)
      *rcode = ncerr;
}

/* takes an open netCDF out of define mode */
void
ncendf_(cdfid, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *rcode;	/* returned error code */
{
    *rcode = 0;
    if (ncendef(*cdfid) == -1)
      *rcode = ncerr;
}

/* returns information about an open netCDF file given its netCDF ID */
void
ncinq_(cdfid, ndims, nvars, natts, recdim, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *ndims;	/* returned number of dimensions */
    int            *nvars;	/* returned number of variables */
    int            *natts;	/* returned number of attributes */
    int            *recdim;	/* returned ID of the unlimited dimension */
    int            *rcode;	/* returned error code */
{
    *rcode = 0;
    if (ncinquire(*cdfid, ndims, nvars, natts, recdim) == -1) {
	*rcode = ncerr;
	return;
    }
    if (*recdim != -1)
      *recdim++;    
}

/* makes sure that the disk copy of a netCDF file open for writing
 * is current
 */
void
ncsnc_(cdfid, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *rcode;	/* returned error code */
{
    *rcode = 0;
    if (ncsync(*cdfid) == -1)
      *rcode = ncerr;
}

/* restores the netCDF to a known consistent state in case anything
 * goes wrong during the definition of new dimensions, variables
 * or attributes
 */
void
ncabor_(cdfid, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *rcode;	/* returned error code */
{
    *rcode = 0;
    if (ncabort(*cdfid) == -1)
      *rcode = ncerr;
}

/* returns the name and size of a dimension, given its ID */
void
ncdinq_(cdfid, dimid, dimname, size, rcode, dimnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *dimid;	/* dimension ID */
    char           *dimname;	/* returned dimension name */
    int            *size;	/* returned dimension size */
    int            *rcode;	/* returned error code */
    int             dimnamelen;	/* declared length of dimname */
{
    char            name[MAX_NC_NAME + 1];
    int             indx, i;

    *rcode = 0;
    if (ncdiminq(*cdfid, *dimid - 1, name, size) == -1) {
	*rcode = ncerr;
	return;
    }
    (void) nstrncpy(dimname, name, dimnamelen);
    indx = strlen(name);
    for (i = indx; i < dimnamelen; i++)
	dimname[i] = ' ';
}

/* renames an existing dimension in a netCDF open for writing */
void
ncdren_(cdfid, dimid, dimname, rcode, dimnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *dimid;	/* dimension ID */
    char           *dimname;	/* new name of dimension */
    int            *rcode;	/* returned error code */
    int             dimnamelen;	/* declared length of dimname */
{
    char            name[MAX_NC_NAME + 1];

    if (dimnamelen > MAX_NC_NAME) {
	*rcode = NC_ENTOOL;
	handle_err("NCDREN", *rcode);
	return;
    }
    (void) nstrncpy(name, dimname, dimnamelen);
    *rcode = 0;
    if (ncdimrename(*cdfid, *dimid - 1, name) == -1)
	*rcode = ncerr;
}

/* returns information about a netCDF variable, given its ID */
void
ncvinq_(cdfid, varid, varname, datatype, ndims, dimarray,
	natts, rcode, varnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *varname;	/* returned variable name */
    int            *datatype;	/* returned variable type */
    int            *ndims;	/* returned number of dimensions */
    int            *dimarray;	/* returned array of ndims dimension IDs */
    int            *natts;	/* returned number of attributes */
    int            *rcode;	/* returned error code */
    int             varnamelen;	/* declared length of varname */
{
    char            name[MAX_NC_NAME + 1];
    int             i, dimid[MAX_VAR_DIMS], indx;

    *rcode = 0;
    if (ncvarinq(*cdfid, *varid - 1, name, (nc_type *) datatype, ndims, dimid,
		 natts) == -1) {
	    *rcode = ncerr;
	    return;
	}
    for (i = 0; i < *ndims; i++)
	dimarray[i] = dimid[i] + 1;
    reverse(dimarray, *ndims);
    if (strlen(name) > varnamelen) {
	*rcode = NC_ESTS;
	handle_err("NCVINQ", *rcode);
	return;
    }
    (void) strcpy(varname, name);
    indx = strlen(name);
    for (i = indx; i < varnamelen; i++)
      varname[i] = ' ';
}

/* puts a single numeric data value into a variable of an open netCDF */
void
ncvpt1_(cdfid, varid, indices, value, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *indices;	/* multidimensional index of data value to be
				   written */
    ncvoid           *value;	/* pointer to data value to be written */
    int            *rcode;	/* returned error code */
{
    int             datatype, ndims, natts, i;
    int             nindices[MAX_VAR_DIMS];
    int             dimid[MAX_VAR_DIMS];

    if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		 (nc_type *) &datatype, &ndims, dimid, &natts) == -1) {
	*rcode = ncerr;
	return;
    }
    for (i = 0; i < ndims; i++)
	nindices[i] = indices[i] - 1;
    reverse(nindices, ndims);
    *rcode = 0;
    if (ncvarput1(*cdfid, *varid - 1, nindices, value) == -1) {
	*rcode = ncerr;
    }
}

/* puts a single character into an open netCDF file */
void
ncvp1c_(cdfid, varid, indices, chval, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *indices;	/* multidimensional index of data value to be
				   written */
    char           *chval;	/* data value to be written */
    int            *rcode;	/* returned error code */
{
    int             datatype, ndims, natts, i;
    int             nindices[MAX_VAR_DIMS], dimid[MAX_VAR_DIMS];

    if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		 (nc_type *) &datatype, &ndims, dimid, &natts) == -1) {
	*rcode = ncerr;
	return;
    }
    for (i = 0; i < ndims; i++)
	nindices[i] = indices[i] - 1;
    reverse(nindices, ndims);
    *rcode = 0;
    if (ncvarput1(*cdfid, *varid - 1, nindices, (ncvoid *) chval) == -1) {
	*rcode = ncerr;
    }
}

/* writes a hypercube of numeric values into a netCDF variable of an open
 * netCDF file
 */
void
ncvpt_(cdfid, varid, start, count, value, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *start;	/* multimensional index of corner of hypercube
				   where first of the data values will be
				   written */
    int            *count;	/* multidimensional edge lengths from the
				   corner of the hypercube where first of
				   data values will be written */
    ncvoid           *value;	/* pointer to block of data values to be
				   written */
    int            *rcode;	/* returned error code */
{
    int             ncount[MAX_VAR_DIMS], nstart[MAX_VAR_DIMS], i;
    int             ndims, datatype, dimarray[MAX_VAR_DIMS], natts;

    if (ncvarinq(*cdfid, *varid - 1, (char *) 0, (nc_type *) &datatype,
		 &ndims, dimarray, &natts) == -1) {
	*rcode = ncerr;
	return;
    }
    for (i = 0; i < ndims; i++) {
	ncount[i] = count[i];
	nstart[i] = start[i] - 1;
    }
    reverse(ncount, ndims);
    reverse(nstart, ndims);
    *rcode = 0;
    if (ncvarput(*cdfid, *varid - 1, nstart, ncount, value) == -1) {
	*rcode = ncerr;
    }
}

/* writes a hypercube of character values into an open netCDF file */
void
ncvptc_(cdfid, varid, start, count, string, lenstr, rcode, stringlen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *start;	/* multidimensional index of corner of
				   hypercube where first of the data values
				   will be written */
    int            *count;	/* multimensional edge lengths from the corner
				   of the hypercube where first of data values
				   will be written */
    char           *string;	/* pointer to block of data values to be
				   written */
    int            *lenstr;	/* declared length of the data argument */
    int            *rcode;	/* returned error code */
    int             stringlen;	/* declared length of string */
{
    int             ncount[MAX_VAR_DIMS], nstart[MAX_VAR_DIMS], i;
    int             ndims, datatype, dimarray[MAX_VAR_DIMS], natts;

    if (*lenstr > stringlen) {
	*rcode = NC_ESTS;
	handle_err("NCVPTC", *rcode);
	return;
    }
    if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		 (nc_type *) &datatype, &ndims, dimarray, &natts) == -1) {
	*rcode = ncerr;
	return;
    }
    for (i = 0; i < ndims; i++) {
	ncount[i] = count[i];
	nstart[i] = start[i] - 1;
    }
    reverse(ncount, ndims);
    reverse(nstart, ndims);
    *rcode = 0;
    if (ncvarput(*cdfid, *varid - 1, nstart, ncount, (ncvoid *) string) == -1) {
	*rcode = ncerr;
    }
}


/* gets a single numeric value from a variable of an open netCDF file */
void
ncvgt1_(cdfid, varid, indices, value, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *indices;	/* multidimensional index of data values to be
				   read */
    ncvoid           *value;	/* pointer to location into which value is
				   read */
    int            *rcode;	/* returned error code */
{
    int             nindices[MAX_VAR_DIMS], i;
    int             datatype, ndims, dimarray[MAX_VAR_DIMS], natts;

    if (ncvarinq(*cdfid, *varid - 1, (char *) 0, (nc_type *) &datatype,
		 &ndims, dimarray, &natts) == -1) {
	*rcode = ncerr;
	return;
    }
    for (i = 0; i < ndims; i++) {
	nindices[i] = indices[i] - 1;
    }
    reverse(nindices, ndims);
    *rcode = 0;
    if (ncvarget1(*cdfid, *varid - 1, nindices, value) == -1) {
	*rcode = ncerr;
    }
}

/* gets a single character data value from a variable of an open
 * netCDF file
 */
void
ncvg1c_(cdfid, varid, indices, chval, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *indices;	/* multidimensional index of the data value to
				   be read */
    char           *chval;	/* location into which data will be read */
    int            *rcode;	/* returned error code */
{
    int             nindices[MAX_VAR_DIMS], i;
    int             datatype, ndims, dimarray[MAX_VAR_DIMS], natts;

    if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		 (nc_type *) &datatype, &ndims, dimarray, &natts) == -1) {
	*rcode = ncerr;
	return;
    }

    for (i = 0; i < ndims; i++) {
	nindices[i] = indices[i] - 1;
    }
    reverse(nindices, ndims);
    *rcode = 0;
    if (ncvarget1(*cdfid, *varid - 1, nindices, (ncvoid *) chval) == -1) {
	*rcode = ncerr;
    }
}

/* reads a hypercube of numeric values from a netCDF variable of an open
 * netCDF file
 */
void
ncvgt_(cdfid, varid, start, count, value, rcode)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *start;	/* multidimensional index of the corner of the
				   hypercube where the first of the data
				   values will be read from */
    int            *count;	/* multidimensional edge lengths from the
				   corner of the hypercube where first of the
				   data values will be read */
    ncvoid           *value;	/* pointer to first of the locations in which
				   the data will be read from the specified
				   hypercube */
    int            *rcode;	/* returned error code */
{
    int             ncount[MAX_VAR_DIMS], nstart[MAX_VAR_DIMS], i;
    int             ndims, datatype, dimarray[MAX_VAR_DIMS], natts;

    if (ncvarinq(*cdfid, *varid - 1, (char *) 0, (nc_type *) &datatype,
		 &ndims, dimarray, &natts) == -1) {
	*rcode = ncerr;
	return;
    }
    for (i = 0; i < ndims; i++) {
	ncount[i] = count[i];
	nstart[i] = start[i] - 1;
    }
    reverse(ncount, ndims);
    reverse(nstart, ndims);
    *rcode = 0;
    if (ncvarget(*cdfid, *varid - 1, nstart, ncount, value) == -1) {
	*rcode = ncerr;
    }
}

/* reads a hypercube of character values from a netCDF variable */
void
ncvgtc_(cdfid, varid, start, count, string, lenstr, rcode, stringlen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *start;	/* multidimensional index of the corner of
				   the hypercube where the first of the data
				   values will be read from */
    int            *count;	/* multidimensional edge lengths from the
				   corner of the hypercube where the first of
				   the data values will be read */
    char           *string;	/* character string into which data values
				   will be read */
    int            *lenstr;	/* declared length of data argument */
    int            *rcode;	/* returned error code */
    int             stringlen;	/* declared length of string */
{
    int             ncount[MAX_VAR_DIMS], nstart[MAX_VAR_DIMS], i;
    int             ndims, datatype, dimarray[MAX_VAR_DIMS], natts;

    if (*lenstr > stringlen) {
	*rcode = NC_ESTS;
	handle_err("NCVGTC", *rcode);
	return;
    }
    if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		 (nc_type *) &datatype, &ndims, dimarray, &natts) == -1) {
	*rcode = ncerr;
	return;
    }
    for (i = 0; i < ndims; i++) {
	ncount[i] = count[i];
	nstart[i] = start[i] - 1;
    }
    reverse(ncount, ndims);
    reverse(nstart, ndims);
    *rcode = 0;
    if (ncvarget(*cdfid, *varid - 1, nstart, ncount, (ncvoid *) string) == -1) {
	*rcode = ncerr;
	return;
    }
    *lenstr = strlen(string);
    for (i = *lenstr; i < stringlen; i++)
	string[i] = ' ';
}

/* changes the name of a netCDF variable in an open netCDF file */
void
ncvren_(cdfid, varid, varname, rcode, varnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *varname;	/* new name for variable */
    int            *rcode;	/* returned error code */
    int             varnamelen;	/* declared length of varname */
{
    char            name[MAX_NC_NAME + 1];

    (void) nstrncpy(name, varname, varnamelen);
    *rcode = 0;
    if (ncvarrename(*cdfid, *varid - 1, name) == -1) {
	*rcode = ncerr;
    }
}

/* adds or changes a numeric variable or global attribute of an open
 * netCDF file
 */
void
ncapt_(cdfid, varid, attname, datatype, attlen, value, rcode,
       attnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *attname;	/* attribute name */
    int            *datatype;	/* attribute datatype */
    int            *attlen;	/* attribute length */
    ncvoid           *value;	/* pointer to data values */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
{
    char            name[MAX_NC_NAME + 1];

    (void) nstrncpy(name, attname, attnamelen);

    *rcode = 0;
    if (ncattput(*cdfid, *varid - 1, name, (nc_type) *datatype, *attlen,
		 (ncvoid *) value) == -1) {
	*rcode = ncerr;
    }
}

/* adds or changes a character variable or global attribute
 * of an open netCDF file
 */
void
ncaptc_(cdfid, varid, attname, datatype, lenstr, string, rcode,
	attnamelen, stringlen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *attname;	/* attribute name */
    int            *datatype;	/* attribute datatype */
    int            *lenstr;	/* length of data argument */
    char           *string;	/* character string value */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
    int             stringlen;	/* declared length of stringlen */
{
    char            name[MAX_NC_NAME + 1];
    char           *value, *malloc();

    if (stringlen > *lenstr) {
	*rcode = NC_ESTS;
	handle_err("NCAPTC", *rcode);
	return;
    }
    (void) nstrncpy(name, attname, attnamelen);
    if (((value = malloc((unsigned) *lenstr + 1)) == NULL) || (*lenstr == 0)) {
	*rcode = NC_ESTS;
	handle_err("NCVPTC", *rcode);
	return;
    }
    fstrncpy(value, string, *lenstr);
    *rcode = 0;
    if (ncattput(*cdfid, *varid - 1, name, (nc_type) *datatype, *lenstr + 1,
		 (ncvoid *) value) == -1) {
	*rcode = ncerr;
    }
}

/* returns information about a netCDF attribute given its variable
 * ID and name
 */
void
ncainq_(cdfid, varid, attname, datatype, attlen, rcode, attnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *attname;	/* attribute name */
    int            *datatype;	/* attribute datatype */
    int            *attlen;	/* attribute length */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
{
    char            name[MAX_NC_NAME + 1];

    (void) nstrncpy(name, attname, attnamelen);
    *rcode = 0;
    if (ncattinq(*cdfid, *varid - 1, name, (nc_type *)datatype, attlen) == -1) {
	*rcode = ncerr;
    }
    if ((nc_type) *datatype == NC_CHAR)
      *attlen = *attlen - 1;
}

/* gets the value of a netCDF attribute given its variable ID
 * and name
 */
void
ncagt_(cdfid, varid, attname, value, rcode, attnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *attname;	/* attribute name */
    ncvoid           *value;	/* returned attribute values */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
{
    char            name[MAX_NC_NAME + 1];

    (void) nstrncpy(name, attname, attnamelen);
    *rcode = 0;
    if (ncattget(*cdfid, *varid - 1, name, value) == -1) {
	*rcode = ncerr;
    }
}

/* gets the value of a netCDF character attribute given its variable
 * ID and name
 */
void
ncagtc_(cdfid, varid, attname, string, lenstr, rcode, attnamelen,
	stringlen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *attname;	/* attribute name */
    char           *string;	/* returned attribute value */
    int            *lenstr;	/* length of string argument */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
{
    char            name[MAX_NC_NAME + 1];
    int             i;

    if (*lenstr > stringlen) {
	*rcode = NC_ESTS;
	handle_err("NCAGTC", *rcode);
	return;
    }
    (void) nstrncpy(name, attname, attnamelen);
    *rcode = 0;
    if (ncattget(*cdfid, *varid - 1, name, (ncvoid *) string) == -1) {
	*rcode = ncerr;
	return;
    }
    *lenstr = strlen(string);
    for (i = *lenstr; i < stringlen; i++)
	string[i] = ' ';
}

/* copies an attribute from one open netCDF file to another */
void
ncacpy_(incdfid, invarid, attname, outcdfid, outvarid, rcode, attnamelen)
    int            *incdfid;	/* netCDF ID */
    int            *invarid;	/* variable ID of input netCDF or NC_GLOBAL */
    char           *attname;	/* name of attribute in input netCDF to be
				   copied */
    int            *outcdfid;	/* netCDF ID of output netCDF file to which
				   attribute will be copied */
    int            *outvarid;	/* ID of the variable in output netCDF file
				   to which attribute will be copied or
				   NC_GLOBAL */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
{
    char            name[MAX_NC_NAME + 1];

    (void) nstrncpy(name, attname, attnamelen);
    *rcode = 0;
    if (ncattcopy(*incdfid, *invarid - 1, name,
		  *outcdfid, *outvarid - 1) == -1) {
	*rcode = ncerr;
    }
}

/* gets the name of an attribute given its variable ID and number
 * as an attribute of that variable
 */
void
ncanam_(cdfid, varid, attnum, attname, rcode, attnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    int            *attnum;	/* attribute number */
    char           *attname;	/* returned attribute name */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
{
    char            name[MAX_NC_NAME + 1];
    int             indx, i;

    *rcode = 0;
    if (ncattname(*cdfid, *varid - 1, *attnum - 1, name) == -1) {
	*rcode = ncerr;
	return;
    }
    (void) nstrncpy(attname, name, attnamelen);
    indx = strlen(name);
    for (i = indx; i < attnamelen; i++)
	attname[i] = ' ';
}


/* renames an attribute in an open netCDF file */
void
ncaren_(cdfid, varid, attname, newname, rcode, attnamelen, newnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *attname;	/* attribute name */
    char           *newname;	/* new name */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
    int             newnamelen;	/* declared length of newname */
{
    char            name[MAX_NC_NAME + 1], nname[MAX_NC_NAME + 1];

    (void) nstrncpy(name, attname, attnamelen);
    (void) nstrncpy(nname, newname, newnamelen);
    *rcode = 0;
    if (ncattrename(*cdfid, *varid - 1, name, nname) == -1) {
	*rcode = ncerr;
    }
}

/* deletes an attribute from an open netCDF file given the attribute
 * name
 */
void
ncadel_(cdfid, varid, attname, rcode, attnamelen)
    int            *cdfid;	/* netCDF ID */
    int            *varid;	/* variable ID */
    char           *attname;	/* attribute name */
    int            *rcode;	/* returned error code */
    int             attnamelen;	/* declared length of attname */
{
    char            name[MAX_NC_NAME + 1];

    (void) nstrncpy(name, attname, attnamelen);
    *rcode = 0;
    if (ncattdel(*cdfid, *varid - 1, name) == -1) {
	*rcode = ncerr;
    }
}


#ifdef NCLINK
/* creates a soft link to a netCDF */
void
nclnk_(path, incdfid, ndims, dimarray, dimmins, dimmaxs, nvars, varids,
       linkcdfid, rcode, pathlen)
    char           *path;	/* path name of input netCDF file */
    int            *incdfid;	/* netCDF ID of open netCDF file to be used as
				   source for variables and data for the link
				   netCDF */
    int            *ndims;	/* number of dimensions of source netCDF
				   required in the link netCDF */
    int            *dimarray;	/* array of dimension IDs of the source
				   netCDF required in the link netCDF */
     int            *dimmins;	/* array of length ndims of dimension-range
				   floors, specifying the bottom of the
				   hypercube subset of data to be used by the
				   netCDF link */
    int            *dimmaxs;	/* array of length ndims of dimension-range
				   ceilings, specifying the top of the
				   hypercube subset of data to be used by
				   the netCDF link */
    int            *nvars;	/* number of variables from the source netCDF
				   file required in the link netCDF file */
    int            *varids;	/* array of length nvars of the variable IDs
				   in the source netCDF file required for the
				   length netCDF file */
    int            *linkcdfid;	/* netCDF ID of the link netCDF file, which
				   should have just been created by a call
				   to nccre */
    int            *rcode;	/* returned error code */
    int             pathlen;	/* declared length of path */
{
    char           *name, *malloc();
    int             nvarids[MAX_NC_VARS], ndimmin[MAX_NC_DIMS];
    int             ndimmax[MAX_NC_DIMS], dimids[MAX_NC_DIMS], i;

    name = malloc((unsigned) pathlen + 1);
    (void) nstrncpy(name, path, pathlen);
    for (i = 0; i < *nvars - 1; i++) {
	nvarids[i] = varids[i] - 1;
    }
    for (i = 0; i < *ndims; i++) {
	dimids[i] = dimarray[i] - 1;
	ndimmin[i] = dimmins[i] - 1;
	ndimmax[i] = dimmaxs[i] - 1;
    }
    reverse(dimarray, *ndims);
    reverse(ndimmin, *ndims);
    reverse(ndimmax, *ndims);
    *rcode = 0;
    if (nclink(name, *incdfid, *ndims, dimids, ndimmin, ndimmax, *nvars,
	       nvarids, *linkcdfid) == -1) {
	*rcode = ncerr;
    }
}

#endif
