/************************************************************************
 *	Copyright 1987, University Corporation for Atmospheric Research *
 *		Not for Resale. All copies to include this notice.      *
 *                                                                      *
 ************************************************************************/

/*-
  OVERVIEW
  This file contains jacket routines written in C for interfacing
  Fortran netCDF function calls to the actual C binding for the
  NetCDF.  This code is written explicitly for VMS.  In general,
  these functions convert between strings represented as
  descriptors and character pointers, convert between column-major-order
  arrays and row-major-order arrays, and map between array indicing
  beginning at one and array indicing beginning at zero.
  -*/


#include	"netcdf.h"
#include	<ctype.h>
#include        <string.h>
#include        ssdef
#include        stdio
#include        descrip

/*
 * global integer used for suppressing error messages and determining the
 * fatality of errors.
 */
extern int      ncopts;		/* default is (NC_FATAL | NC_VERBOSE) */

/* global integer that contains a netCDF-specific error code */
extern int      ncerr;

/* VMS data structure for storing strings */
typedef struct dsc$descriptor_s *dscrp;


/* used to set the C global variable ncopts from Fortran */
int
ncpopt(val)
    int            *val;  /* NC_FATAL, NC_VERBOSE, or NC_FATAL|NC_VERBOSE */
{
    ncopts = *val;
}


/* used to get the C global variable ncopts from Fortran */
int
ncgopt(val)
    int            *val;  /* NC_FATAL, NC_VERBOSE, or NC_FATAL|NC_VERBOSE */
{
    *val = ncopts;
}


static void
reverse(array, length)
     int             array[];	/* array to be reversed */
     int             length;	/* length of array */
{
     int             temp, i, j;
     for (i = 0, j = length - 1; i < j; i++, j--) {
	  temp = array[i];
	  array[i] = array[j];
	  array[j] = temp;
     }
}


/* error handling function */
static void
handle_err(pname, rcode)
     char           *pname;	/* procedure name */
     int             rcode;	/* error return */
{
     if (rcode == 0)
	  return;
     if (ncopts & NC_VERBOSE) {
	  switch (rcode) {
	  case NC_ESTS:
	       (void) fprintf(stderr,
			      "%s: In FORTRAN interface, string too small\n",
			      pname);
	       break;
	  case NC_ENTOOL:
	       (void) fprintf(stderr,
			      "%s: length of name exceeds MAX_NC_NAME\n",
			      pname);
	       break;
	  default:
	       (void) fprintf(stderr,
			      "%s: unknown error code\n", pname);
	  }
     }
     if (ncopts & NC_FATAL)
	  exit(1);
}

/*
 * function to blank fill fortran descriptors before handing back to fortran
 */
static void
blnkfil(target, maxlen)
     char           *target;	/* string to be blank filled */
     int             maxlen;	/* length of string */
{
     int             indx, i;
     indx = strlen(target);
     for (i = indx; i < maxlen; i++)
	  target[i] = ' ';
}

/* copy function used to copy strings with embedded blanks */
static void
fstrncpy(target, source, maxlen)
     char           *target;	/* space to be copied into */
     char           *source;	/* string to be copied */
     int             maxlen;	/* maximum length of *source */
{
     while (maxlen-- && isprint(*source))
	  *target++ = *source++;
     *target = '\0';
}

/* copy function used to copy strings terminated with blanks */
static void
nstrncpy(target, source, maxlen)
     char           *target;	/* space to be copied into */
     char           *source;	/* string to be copied */
     int             maxlen;	/* maximum length of *source */
{
     while (maxlen-- && isgraph(*source))
	  *target++ = *source++;
     *target = '\0';
}


/* ------------ VMS FORTRAN jackets for netCDF Functions ------------ */

/*
 * creates a new netCDF file, returning a netCDF ID.  New netCDF file is
 * placed in define mode.
 */
int
nccre(pathname, clobmode, rcode)
     dscrp           pathname;	/* file name of new netCDF file */
     int            *clobmode;	/* either NCCLOB or NCNOCLOB */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];
     int             cdfid;

     (void) fstrncpy(name, pathname->dsc$a_pointer, pathname->dsc$w_length);
     if ((cdfid = nccreate(name, *clobmode)) != -1) {
	  *rcode = 0;
	  return (cdfid);
     }
     *rcode = ncerr;
     return (-1);
}


/* opens an existing netCDF file for access */
int
ncopn(pathname, rwmode, rcode)
     dscrp           pathname;	/* file name for netCDF to be opened */
     int            *rwmode;	/* either NCWRITE or NCNOWRIT */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];
     int             cdfid;

     (void) nstrncpy(name, pathname->dsc$a_pointer, pathname->dsc$w_length);
     if ((cdfid = ncopen(name, *rwmode)) != -1) {
	  *rcode = 0;
	  return (cdfid);
     }
     *rcode = ncerr;
     return (-1);
}


/* adds a new dimension to an open netCDF file in define mode */
int
ncddef(cdfid, dimname, dimlen, rcode)
     int            *cdfid;	/* netCDF ID */
     dscrp           dimname;	/* Dimension name */
     int            *dimlen;	/* size of dimension */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];
     int             dimid;
     (void) nstrncpy(name, dimname->dsc$a_pointer, dimname->dsc$w_length);
     if ((dimid = ncdimdef(*cdfid, name, *dimlen)) != -1) {
	  *rcode = 0;
	  return (dimid + 1);
     }
     *rcode = ncerr;
     return (-1);
}

/*
 * returns the ID of a netCDF dimension, given the name of the dimension
 */
int
ncdid(cdfid, dimname, rcode)
     int            *cdfid;	/* netCDF ID * dscrp dimname; * dimension
				 * name */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];
     int             dimid;
     (void) nstrncpy(name, dimname->dsc$a_pointer, dimname->dsc$w_length);
     if ((dimid = ncdimid(*cdfid, name)) != -1) {
	  *rcode = 0;
	  return (dimid + 1);
     }
     *rcode = ncerr;
     return (-1);
}


/* adds a new variable to an open netCDF file in define mode */
int
ncvdef(cdfid, varname, datatype, ndims, dimarray, rcode)
     int            *cdfid;	/* netCDF ID */
     dscrp           varname;	/* name of variable */
     int            *datatype;	/* netCDF datatype of variable */
     int            *ndims;	/* number of dimensions of variable */
     int            *dimarray;	/* array of ndims dimensions IDs */
     int            *rcode;	/* returned error code */
{
     int             varid, i, dimid[MAX_VAR_DIMS];
     char            name[MAX_NC_NAME + 1];
     (void) nstrncpy(name, varname->dsc$a_pointer, varname->dsc$w_length);
     for (i = 0; i < *ndims; i++)
	  dimid[i] = dimarray[i] - 1;
     reverse(dimid, *ndims);
     if ((varid = ncvardef(*cdfid, name, (nc_type) (nc_type) * datatype,
			   *ndims, dimid)) != -1) {
	  *rcode = 0;
	  return (varid + 1);
     }
     *rcode = ncerr;
     return (-1);
}

/* returns the ID of a netCDF variable given its name */
int
ncvid(cdfid, varname, rcode)
     int            *cdfid;	/* netCDF ID */
     dscrp           varname;	/* variable name */
     int            *rcode;	/* returned error code */
{
     int             varid;
     char            name[MAX_NC_NAME + 1];
     (void) nstrncpy(name, varname->dsc$a_pointer, varname->dsc$w_length);
     if ((varid = ncvarid(*cdfid, name)) != -1) {
	  *rcode = 0;
	  return (varid + 1);
     }
     *rcode = ncerr;
     return (-1);
}



/* returns number of bytes per netCDF data type */
int
nctlen(datatype, rcode)
     int            *datatype;	/* netCDF datatype */
     int            *rcode;	/* returned error code */
{
     int             itype;
     if ((itype = nctypelen((nc_type) * datatype)) != -1) {
	  *rcode = 0;
	  return (itype);
     }
     *rcode = ncerr;
     return (-1);
}



/* ------------ VMS FORTRAN jackets for netCDF Subroutines ---------- */



/* closes an open netCDF file */
void
ncclos(cdfid, rcode)
     int            *cdfid;	/* netCDf ID */
     int            *rcode;	/* returned error code */
{
     *rcode = 0;
     if (ncclose(*cdfid) == -1)
	  *rcode = ncerr;
}

/* puts an open netCDF into define mode */
void
ncredf(cdfid, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *rcode;	/* returned error code */
{
     *rcode = 0;
     if (ncredef(*cdfid) == -1)
	  *rcode = ncerr;
}

/* takes an open netCDF out of define mode */
void
ncendf(cdfid, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *rcode;	/* returned error code */
{
     *rcode = 0;
     if (ncendef(*cdfid) == -1)
	  *rcode = ncerr;
}

/* returns information about an open netCDF file given its netCDF ID */
void
ncinq(cdfid, ndims, nvars, natts, recdim, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *ndims;	/* returned number of dimensions */
     int            *nvars;	/* returned number of variables */
     int            *natts;	/* returned number of attributes */
     int            *recdim;	/* returned ID of the unlimited dimension */
     int            *rcode;	/* returned error code */
{
     *rcode = 0;
     if (ncinquire(*cdfid, ndims, nvars, natts, recdim) == -1) {
	*rcode = ncerr;
	return;
    }
    if (*recdim != -1)
      *recdim++;    
}

/*
 * makes sure that the disk copy of a netCDF file open for writing is current
 */
void
ncsnc(cdfid, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *rcode;	/* returned error code */
{
     *rcode = 0;
     if (ncsync(*cdfid) == -1)
	  *rcode = ncerr;
}

/*
 * restores the netCDF to a known consistent state in case anything goes
 * wrong during the definition of new dimensions, variables or attributes
 */
void
ncabor(cdfid, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *rcode;	/* returned error code */
{
     *rcode = 0;
     if (ncabort(*cdfid) == -1)
	  *rcode = ncerr;
}

/* returns the name and size of a dimension, given its ID */
void
ncdinq(cdfid, dimid, dimname, size, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *dimid;	/* dimension id */
     dscrp           dimname;	/* returned dimension name */
     int            *size;	/* returned dimension size */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];

     *rcode = 0;
     if (ncdiminq(*cdfid, *dimid - 1, name, size) == -1) {
	  *rcode = ncerr;
	  return;
     }
     if (strlen(name) > dimname->dsc$w_length) {
	  *rcode = NC_ESTS;
	  handle_err("NCDINQ", *rcode);
	  return;
     }
     blnkfil(name, dimname->dsc$w_length);
     /* create a descriptor for returned dimension name */
     (void) strncpy(dimname->dsc$a_pointer, name, dimname->dsc$w_length);
     dimname->dsc$b_class = DSC$K_CLASS_S;
     dimname->dsc$b_dtype = DSC$K_DTYPE_T;
}

/* renames an existing dimension in a netCDF open for writing */
void
ncdren(cdfid, dimid, dimname, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *dimid;	/* dimension id */
     dscrp           dimname;	/* new name of dimension */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];
     if (dimname->dsc$w_length > MAX_NC_NAME) {
	  *rcode = NC_ENTOOL;
	  handle_err("NCDREN", *rcode);
	  return;
     }
     (void) nstrncpy(name, dimname->dsc$a_pointer, dimname->dsc$w_length);
     *rcode = 0;
     if (ncdimrename(*cdfid, *dimid - 1, name) == -1)
	  *rcode = ncerr;
}


/* returns information about a netCDF variable, given its ID */
void
ncvinq(cdfid, varid, varname, datatype, ndims, dimarray,
       natts, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           varname;	/* returned variable name */
     int            *datatype;	/* returned variable type */
     int            *ndims;	/* returned number of dimensions */
     int            *dimarray;	/* returned array of ndims dimension IDs */
     int            *natts;	/* returned number of attributes */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];
     int             dimid[MAX_VAR_DIMS], i;

     *rcode = 0;
     if (ncvarinq(*cdfid, *varid - 1, name, (nc_type *) datatype, ndims, dimid,
		  natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < *ndims; i++)
	  dimarray[i] = dimid[i] + 1;
     reverse(dimarray, *ndims);
     if (strlen(name) > varname->dsc$w_length) {
	  *rcode = NC_ESTS;
	  handle_err("NCVINQ", *rcode);
	  return;
     }
     blnkfil(name, varname->dsc$w_length);
     /* create a descriptor */
     (void) strncpy(varname->dsc$a_pointer, name, varname->dsc$w_length);
     varname->dsc$b_class = DSC$K_CLASS_S;
     varname->dsc$b_dtype = DSC$K_DTYPE_T;
}

/* puts a single numeric data value into a variable of an open netCDF */
void
ncvpt1(cdfid, varid, indices, value, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *indices;	/* multidimensional index of data 
                                   value to be written */
     void           *value;	/* pointer to data value to be written */
     int            *rcode;	/* returned error code */
{
     int             datatype, ndims, natts, i;
     int             nindices[MAX_VAR_DIMS], dimid[MAX_VAR_DIMS];
     if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		  (nc_type *) & datatype, &ndims, dimid, &natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < ndims; i++)
	  nindices[i] = indices[i] - 1;
     reverse(nindices, ndims);
     *rcode = 0;
     if (ncvarput1(*cdfid, *varid - 1, nindices, value) == -1) {
	  *rcode = ncerr;
     }
}

/* puts a single character into an open netCDF file */
void
ncvp1c(cdfid, varid, indices, chval, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *indices;	/* multidimensional index of data 
                                  value to be written */
     char           *chval;	/* data value to be written */
     int            *rcode;	/* returned error code */
{
     int             datatype, ndims, natts, i;
     int             nindices[MAX_VAR_DIMS], dimid[MAX_VAR_DIMS];

     if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		  (nc_type *) & datatype, &ndims, dimid, &natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < ndims; i++)
	  nindices[i] = indices[i] - 1;
     reverse(nindices, ndims);
     *rcode = 0;
     if (ncvarput1(*cdfid, *varid - 1, nindices, (void *) chval) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * writes a hypercube of numeric values into a netCDF variable of an open
 * netCDF file
 */
void
ncvpt(cdfid, varid, start, counts, value, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *start;	/* multidimensional index of corner 
                                   of hypercube where first of the 
                                   data values will be written */
     int            *counts;	/* multidimensional edge lengths 
                                   from the corner of the hypercube 
                                   where first of data values will 
                                   be written */
     void           *value;	/* pointer to block of data values 
                                   to be written */
     int            *rcode;	/* returned error code */
{
     int             ncount[MAX_VAR_DIMS], nstart[MAX_VAR_DIMS], i;
     int             ndims, datatype, dimarray[MAX_VAR_DIMS], natts;

     if (ncvarinq(*cdfid, *varid - 1, (char *) 0, (nc_type *) & datatype,
		  &ndims, dimarray, &natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < ndims; i++) {
	  ncount[i] = counts[i];
	  nstart[i] = start[i] - 1;
     }
     reverse(ncount, ndims);
     reverse(nstart, ndims);
     *rcode = 0;
     if (ncvarput(*cdfid, *varid - 1, nstart, ncount, value) == -1) {
	  *rcode = ncerr;
     }
}

/* writes a hypercube of character values into an open netCDF file */
void
ncvptc(cdfid, varid, start, counts, string, lenstr, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *start;	/* multidimensional index of corner 
				   of hypercube where first of the 
                                   data values will be written */
     int            *counts;	/* multidimensional edge lengths 
                                   from the corner of the hypercube 
                                   where first of data values will 
                                   be written */
     dscrp           string;	/* pointer to block of data values 
                                   to be written */
     int            *lenstr;	/* declared length of the data argument */
     int            *rcode;	/* returned error code */
{
     int             ncount[MAX_VAR_DIMS], nstart[MAX_VAR_DIMS], i;
     int             ndims, datatype, dimarray[MAX_VAR_DIMS], natts;
     char           *value, *malloc();

     if (*lenstr > string->dsc$w_length) {
	  *rcode = NC_ESTS;
	  handle_err("NCVPTC", *rcode);
	  return;
     }
     if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		  (nc_type *) & datatype, &ndims, dimarray, &natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < ndims; i++) {
	  ncount[i] = counts[i];
	  nstart[i] = start[i] - 1;
     }
     reverse(ncount, ndims);
     reverse(nstart, ndims);
     if (((value = malloc((unsigned) *lenstr + 1)) == NULL)
	 || (*lenstr == 0)) {
	  *rcode = NC_ESTS;
	  handle_err("NCVPTC", *rcode);
	  return;
     }
     (void) fstrncpy(value, string->dsc$a_pointer, string->dsc$w_length);
     *rcode = 0;
     if (ncvarput(*cdfid, *varid - 1, nstart, ncount, (void *) value) == -1) {
	  *rcode = ncerr;
     }
}


/* gets a single numeric value from a variable of an open netCDF file */
void
ncvgt1(cdfid, varid, indices, value, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *indices;	/* multidimensional index of data values 
                                   to be read */
     void           *value;	/* pointer to location into which 
                                   data value is read */
     int            *rcode;	/* returned error code */
{
     int             nindices[MAX_VAR_DIMS], i;
     int             datatype, ndims, dimarray[MAX_VAR_DIMS], natts;
     if (ncvarinq(*cdfid, *varid - 1, (char *) 0, (nc_type *) & datatype,
		  &ndims, dimarray, &natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < ndims; i++) {
	  nindices[i] = indices[i] - 1;
     }
     reverse(nindices, ndims);
     *rcode = 0;
     if (ncvarget1(*cdfid, *varid - 1, nindices, value) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * gets a single character data value from a variable of an open netCDF file
 */
void
ncvg1c(cdfid, varid, indices, chval, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *indices;	/* multidimensional index of the 
                                   data value to be read */
     char           *chval;	/* location into which data 
                                   will be read */
     int            *rcode;	/* returned error code */
{
     int             nindices[MAX_VAR_DIMS], i;
     int             datatype, ndims, dimarray[MAX_VAR_DIMS], natts;
     if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		  (nc_type *) & datatype, &ndims, dimarray, &natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < ndims; i++) {
	  nindices[i] = indices[i] - 1;
     }
     reverse(nindices, ndims);
     *rcode = 0;
     if (ncvarget1(*cdfid, *varid - 1, nindices, (void *) chval) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * reads a hypercube of numeric values from a netCDF variable of an open
 * netCDF file
 */
void
ncvgt(cdfid, varid, start, count, value, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *start;	/* multidimensional index of the 
                                   corner of the hypercube where 
                                   the first of the data values  
                                   will be read from */
     int            *count;	/* multidimensional edge lengths 
                                   from the corner of the hypercube 
                                   where first of the data values 
                                   will be read */
     void           *value;	/* pointer to first of the locations 
                                   in which the data will be read 
                                   from the specified hypercube */
     int            *rcode;	/* returned error code */
{
     int             ncount[MAX_VAR_DIMS], nstart[MAX_VAR_DIMS], i;
     int             ndims, datatype, dimarray[MAX_VAR_DIMS], natts;
     if (ncvarinq(*cdfid, *varid - 1, (char *) 0, (nc_type *) & datatype,
		  &ndims, dimarray, &natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < ndims; i++) {
	  ncount[i] = count[i];
	  nstart[i] = start[i] - 1;
     }
     reverse(ncount, ndims);
     reverse(nstart, ndims);
     *rcode = 0;
     if (ncvarget(*cdfid, *varid - 1, nstart, ncount, value) == -1) {
	  *rcode = ncerr;
     }
}

/* reads a hypercube of character values from a netCDF variable */
void
ncvgtc(cdfid, varid, start, count, string, lenstr, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *start;	/* multidimensional index of the 
                                   corner of the hypercube where 
                                   the first of the data values 
                                   will be read from */
     int            *count;	/* multidimensional edge lengths 
                                   from the corner of the hypercube 
                                   where the first of the data 
                                   values will be read */
     dscrp           string;	/* character string into which data 
                                   will be read */
     int            *lenstr;	/* declared length of data argument */
     int            *rcode;	/* returned error code */
{
     int             ncount[MAX_VAR_DIMS], nstart[MAX_VAR_DIMS], i;
     int             ndims, datatype, dimarray[MAX_VAR_DIMS], natts;
     char           *value, *malloc();
     if (*lenstr > string->dsc$w_length) {
	  *rcode = NC_ESTS;
	  handle_err("NCVGTC", *rcode);
	  return;
     }
     if (ncvarinq(*cdfid, *varid - 1, (char *) 0,
		  (nc_type *) & datatype, &ndims, dimarray, &natts) == -1) {
	  *rcode = ncerr;
	  return;
     }
     for (i = 0; i < ndims; i++) {
	  ncount[i] = count[i];
	  nstart[i] = start[i] - 1;
     }
     reverse(ncount, ndims);
     reverse(nstart, ndims);
     if (((value = malloc((unsigned) *lenstr + 1)) == NULL)
	 || (*lenstr == 0)) {
	  *rcode = NC_ESTS;
	  handle_err("NCVGTC", *rcode);
	  return;
     }
     *rcode = 0;
     if (ncvarget(*cdfid, *varid - 1, nstart, ncount, value) == -1) {
	  *rcode = ncerr;
	  return;
     }
     /* create a descriptor  */
     blnkfil(value, string->dsc$w_length);
     (void) strncpy(string->dsc$a_pointer, value, string->dsc$w_length);
     string->dsc$b_class = DSC$K_CLASS_S;
     string->dsc$b_dtype = DSC$K_DTYPE_T;
}

/* changes the name of a netCDF variable in an open netCDF file */
void
ncvren(cdfid, varid, varname, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           varname;	/* new name for variable */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];

     (void) nstrncpy(name, varname->dsc$a_pointer, varname->dsc$w_length);
     *rcode = 0;
     if (ncvarrename(*cdfid, *varid - 1, name) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * adds or changes a numeric variable or global attribute of an open netCDF
 * file
 */
void
ncapt(cdfid, varid, attname, datatype, attlen, value, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           attname;	/* attribute name */
     int            *datatype;	/* attribute datatype */
     int            *attlen;	/* attribute length */
     void           *value;	/* pointer to data values */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];
     nstrncpy(name, attname->dsc$a_pointer, attname->dsc$w_length);

     *rcode = 0;
     if (ncattput(*cdfid, *varid - 1, name, (nc_type) * datatype,
		  *attlen, value) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * adds or changes a character variable or global attribute of an open netCDF
 * file
 */
void
ncaptc(cdfid, varid, attname, datatype, lenstr, string, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           attname;	/* attribute name */
     int            *datatype;	/* attribute datatype */
     int            *lenstr;	/* length of data argument */
     dscrp           string;	/* character string value */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];
     char           *value, *malloc();

     (void) nstrncpy(name, attname->dsc$a_pointer, attname->dsc$w_length);
     if (((value = malloc((unsigned) *lenstr + 1)) == NULL) || (*lenstr == 0)) {
	  *rcode = NC_ESTS;
	  handle_err("NCAPTC", *rcode);
	  return;
     }
     (void) fstrncpy(value, string->dsc$a_pointer, string->dsc$w_length);
     *rcode = 0;
     if (ncattput(*cdfid, *varid - 1, name, (nc_type) * datatype, *lenstr + 1,
		  (void *) value) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * returns information about a netCDF attribute given its variable ID and
 * name
 */
void
ncainq(cdfid, varid, attname, datatype, attlen, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           attname;	/* returned attribute name */
     int            *datatype;	/* returned attribute datatype */
     int            *attlen;	/* returned attribute length */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];

     (void) nstrncpy(name, attname->dsc$a_pointer, attname->dsc$w_length);
     *rcode = 0;
     if (ncattinq(*cdfid, *varid - 1, name, (nc_type *) datatype, attlen) == -1) {
	  *rcode = ncerr;
     }
    if ((nc_type) *datatype == NC_CHAR)
       *attlen = *attlen - 1;
}

/*
 * gets the value of a netCDF attribute given its variable ID and name
 */
void
ncagt(cdfid, varid, attname, value, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           attname;	/* attribute name */
     void           *value;	/* returned attribute values */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];

     (void) nstrncpy(name, attname->dsc$a_pointer, attname->dsc$w_length);
     *rcode = 0;
     if (ncattget(*cdfid, *varid - 1, name, value) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * gets the value of a netCDF character attribute given its variable ID and
 * name
 */
void
ncagtc(cdfid, varid, attname, string, lenstr, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           attname;	/* attribute name */
     dscrp           string;	/* returned attribute value */
     int            *lenstr;	/* length of string argument */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1], *value, *malloc();

     (void) nstrncpy(name, attname->dsc$a_pointer, attname->dsc$w_length);
     if (((value = malloc((unsigned) *lenstr + 1)) == NULL)
	 || (*lenstr == 0)) {
	  *rcode = NC_ESTS;
	  handle_err("NCAGTC", *rcode);
	  return;
     }
     *rcode = 0;
     if (ncattget(*cdfid, *varid - 1, name, (void *) value) == -1) {
	  *rcode = ncerr;
	  return;
     }
     blnkfil(value, string->dsc$w_length);
     /* create a descriptor  */
     (void) strncpy(string->dsc$a_pointer, value, string->dsc$w_length);
     string->dsc$b_class = DSC$K_CLASS_S;
     string->dsc$b_dtype = DSC$K_DTYPE_T;
}

/* copies an attribute from one open netCDF file to another */
void
ncacpy(incdfid, invarid, attname, outcdfid, outvarid, rcode)
     int            *incdfid;	/* input netCDF ID */
     int            *invarid;	/* variable ID of input netCDF 
                                   or NC_GLOBAL */
     dscrp           attname;	/* name of attribute in input 
                                   netCDF to be copied */
     int            *outcdfid;	/* netCDF ID of output netCDF 
                                   file to which attribute will 
                                   be copied */
     int            *outvarid;	/* ID of the variable in output 
                                   netCDF file to which attribute 
                                   will be copied or NC_GLOBAL */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];

     (void) nstrncpy(name, attname->dsc$a_pointer, attname->dsc$w_length);
     *rcode = 0;
     if (ncattcopy(*incdfid, *invarid - 1, name, *outcdfid, 
		   *outvarid - 1) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * gets the name of an attribute given its variable ID and number as an
 * attribute of that variable
 */
void
ncanam(cdfid, varid, attnum, attname, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     int            *attnum;	/* attribute number */
     dscrp           attname;	/* returned attribute name */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];

     *rcode = 0;
     if (ncattname(*cdfid, *varid - 1, *attnum - 1, name) == -1) {
	  *rcode = ncerr;
	  return;
     }
     blnkfil(name, attname->dsc$w_length);
     /* create a descriptor  */
     (void) strncpy(attname->dsc$a_pointer, name, attname->dsc$w_length);
     attname->dsc$b_class = DSC$K_CLASS_S;
     attname->dsc$b_dtype = DSC$K_DTYPE_T;
}


/* renames an attribute in an open netCDF file */
void
ncaren(cdfid, varid, attname, newname, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           attname;	/* attribute name */
     dscrp           newname;	/* new name */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1], nname[MAX_NC_NAME + 1];
     (void) nstrncpy(name, attname->dsc$a_pointer, attname->dsc$w_length);
     (void) nstrncpy(nname, newname->dsc$a_pointer, newname->dsc$w_length);
     *rcode = 0;
     if (ncattrename(*cdfid, *varid - 1, name, nname) == -1) {
	  *rcode = ncerr;
     }
}

/*
 * deletes an attribute from an open netCDF file given the attribute name
 */
void
ncadel(cdfid, varid, attname, rcode)
     int            *cdfid;	/* netCDF ID */
     int            *varid;	/* variable ID */
     dscrp           attname;	/* attribute name */
     int            *rcode;	/* returned error code */
{
     char            name[MAX_NC_NAME + 1];

     (void) nstrncpy(name, attname->dsc$a_pointer, attname->dsc$w_length);
     *rcode = 0;
     if (ncattdel(*cdfid, *varid - 1, name) == -1) {
	  *rcode = ncerr;
     }
}


/* creates a soft link to a netCDF */
#ifdef NCLNK
void
nclnk(path, incdfid, ndims, dimarray, dimmins, dimmaxs, nvars, varids,
      linkcdfid, rcode)
     dscrp           path;	/* path name of input netCDF file */
     int            *incdfid;	/* netCDF ID of open netCDF file to be used
				   as source for variables and data for the
				   link netCDF */
     int            *ndims;	/* number of dimensions of source netCDF
				   required in the link netCDF */
     int            *dimarray;	/* array of dimension IDs of the source
				   netCDF required in the link netCDF */
     int            *dimmins;	/* array of length ndims of dimension- range
				   floors, specifying the bottom of the
				   hypercube subset of data to be used by the
				   netCDF 
                                   link           */
     int            *dimmaxs;	/* array of length ndims of dimension- range
				   ceilings, specifying the top of the
				   hypercube subset of data to be used by the
				   netCDF link */
     int            *nvars;	/* number of variables from the source netCDF
				   file required in the link netCDF file */
     int            *varids;	/* array of length nvars of the variable */
     IDs in the source netCDF file required
     for             the length netCDF file * /
     int            *linkcdfid;	/* netCDF ID of the link netCDF file, which
				   should have just been created by a call to
				   nccre */
     int            *rcode;	/* returned error code */
     {
	  char           *name, *malloc();
	  int             nvarids[MAX_NC_VARS], ndimmin[MAX_NC_DIMS];
	  int             ndimmax[MAX_NC_DIMS], dimids[MAX_NC_DIMS], i;
	  name = malloc((unsigned) path->dsc$w_length + 1);
	  fstrncpy(name, path->dsc$a_pointer, path->dsc$w_length);
	  for (i = 0; i < *nvars - 1; i++) {
	       nvarids[i] = varids[i]--;
	  }
	  for (i = 0; i < *ndims - 1; i++) {
	       dimids[i] = dimarray[i]--;
	       ndimmin[i] = dimmins[i]--;
	       ndimmax[i] = dimmaxs[i]--;
	  }
	  (void) reverse(dimids, *ndims);
	  (void) reverse(ndimmin, *ndims);
	  (void) reverse(ndimmax, *ndims);
	  *rcode = 0;
	  if (nclink(name, *incdfid, *ndims, dimids, ndimmin, ndimmax, *nvars,
		     nvarids, *linkcdfid) == -1) {
	       *rcode = ncerr;
	  }
     }

#endif
