/*********************************************************************
 *   Copyright 1989, University Corporation for Atmospheric Research
 *   See netcdf/README file for copying and redistribution conditions.
 *   $Header: /u2/ldm/src/netcdf/nctest/RCS/varput.c,v 1.1 89/05/08 16:54:51 russ Exp $
 *********************************************************************/

#include <stdio.h>
#include "netcdf.h"
#include "ncvoid.h"
#include "testcdf.h"		/* defines in-memory test cdf structure */

extern struct netcdf test;	/* memory netcdf kept sync with disk netcdf */

extern void error();		/* prints error messages */

/*
 * Test ncvarput
 *    check that proper call worked with ncvarget
 *    try with negative coords, edges, check error
 *    try with too-large coords, edges, check error
 *    try with bad variable handle, check error
 *    try in define mode, check error
 *    try with bad netCDF handle, check error
 */
void
test_ncvarput(path)
     char *path;		/* name of writable netcdf file to open */
{
    static char pname[] = "test_ncvarput";
    int cdfid;			/* netcdf id */
    extern void test_varputget(); /* tests both ncvarput and ncvarget */
    int iv;			/* variable id */
    struct cdfhc {		/* a hypercube with generic values */
	int cor[MAX_NC_DIMS];	/* netcdf coordinates for lower corner */
	int edg[MAX_NC_DIMS];	/* netcdf edge lengths to upper corner */
	Void *vals;		/* pointer to block of values */
    } hc;			/* test hypercube */
    extern Void *emalloc();	/* error-checking allocator */
    void val_fill();		/* fills generic block with typed data */
    int tmp;
    int id;			/* dimension id */

    (void) fprintf(stderr, "--- Testing %s ...\n", &pname[5]);

    if ((cdfid = ncopen(path, NC_WRITE)) == -1) {
	error("%s: ncopen failed", pname);
	return;
    }

    /* opened in data mode, try putting and getting hypercubes of each type */
    test_varputget (cdfid);	/* prints out messages for discrepencies */
    
    /* find a variable with at least one dimension */
    iv = 0;
    while (test.vars[iv].ndims <= 0 && iv < test.nvars)
      iv++;
    if (iv < test.nvars) {	/* iv is first varid of var with dimensions */
	/* set coords */
	for (id = 0; id < test.vars[iv].ndims; id++) {
	    hc.cor[id] = 0;
	    hc.edg[id] = 1;
	}
	/* fill in vals with value of appropriate type */
	hc.vals = emalloc(nctypelen(test.vars[iv].type));
	val_fill(test.vars[iv].type, 1, hc.vals);

	id = test.vars[iv].ndims - 1;
	tmp = hc.cor[id];
	hc.cor[id] = -1;	/* try negative coordinate, should fail */
	if(ncvarput (cdfid, iv, hc.cor, hc.edg, hc.vals) != -1) {
	    error("%s: ncvarput should fail for negative corner", pname);
	    ncclose(cdfid); return;
	}
	hc.cor[id] = tmp;
	tmp = hc.edg[id];
	hc.edg[id] = -1;	/* try negative edge, should fail */
	if(ncvarput (cdfid, iv, hc.cor, hc.edg, hc.vals) != -1) {
	    error("%s: ncvarput should fail for negative edge", pname);
	    ncclose(cdfid); return;
	}
	hc.edg[id] = tmp;
	tmp = hc.cor[id];
	hc.cor[id] = test.dims[id].size; /* try big coordinate, should fail */
	if(ncvarput (cdfid, iv, hc.cor, hc.edg, hc.vals) != -1) {
	    error("%s: ncvarput should fail for too-high coordinate", pname);
	    ncclose(cdfid); return;
	}
	hc.cor[id] = tmp;
	tmp = hc.edg[id];
	hc.edg[id] = test.dims[id].size + 1; /* try big edge, should fail */
	if(ncvarput (cdfid, iv, hc.cor, hc.edg, hc.vals) != -1) {
	    error("%s: ncvarput should fail for too-high edge", pname);
	    ncclose(cdfid); return;
	}
	hc.edg[id] = tmp;

	if (ncredef(cdfid) == -1) {
	    error("%s: ncredef failed", pname);
	    ncclose(cdfid); return;
	}
	/* try in define mode, should fail */
	if(ncvarput (cdfid, iv, hc.cor, hc.edg, hc.vals) != -1) {
	    error("%s: ncvarput should fail in define mode", pname);
	    ncclose(cdfid); return;
	}
	if (ncendef (cdfid) == -1) {
	    error("%s: ncendef failed", pname);
	    ncclose(cdfid); return;
	}
    }
    /* try with bad variable handle, should fail */
    if(ncvarput (cdfid, -1, hc.cor, hc.edg, hc.vals) != -1) {
	error("%s: ncvarput should fail for bad variable handle", pname);
	ncclose(cdfid); return;
    }
    if (ncclose (cdfid) == -1) {
	error("%s: ncclose failed", pname);
	return;
    }
    /* try with bad netCDF handle, should fail */
    if(ncvarput (cdfid, 0, hc.cor, hc.edg, hc.vals) != -1)
	error("%s: ncvarput failed to report bad netcdf handle", pname);
    free ((char *)hc.vals);
}
